//
// ERGOT : The Windows MacTierra clone
// copyright (c) 2005 plushpuffin@gmail.com
// http://www.wwddfd.com/
//
// latest build: http://www.wwddfd.com/frell/ergot.zip
//
// licensing terms still under consideration...

using System;
using System.Xml;

namespace ERGOT
{
	/// <summary>
	/// An ergot is actually a thread which executes whatever instructions it happens to pass over in memory.
	/// Ancestral ergots pretty much stick to their own memory space, but deliberate, random flaws in execution,
	/// copying, and random mutations in memory will cause all kinds of different schemes for reproduction,
	/// including parasitism (loss of the creature's own reproduction instructions, and use of another creature's
	/// instruction set to copy itself more quickly)
	/// </summary>
	public class ergot
	{
		#region Member Constants
		public const int c_stacksize = 10;
		static readonly byte[] c_spdec = new byte[] {9,0,1,2,3,4,5,6,7,8};
		static readonly byte[] c_spinc = new byte[] {1,2,3,4,5,6,7,8,9,0};
		#endregion

		#region Member Variables
		///////////////////////////////////////////////////////////////////
		///////////////////////////////////////////////////////////////////
		/// my registers...
		public short	ax;			//Address register
		public short	bx;			//Address register
		public short	cx;			//Numerical register
		public short	dx;			//Numerical register
		public bool		fl;			//Flag (set when errors occur)
		public byte		sp;			//Stack pointer
		public short[]	st;			//Stack
		public short	ip;			//Instruction pointer
		///////////////////////////////////////////////////////////////////
		///////////////////////////////////////////////////////////////////
		/// my variables...
		private int myPosition;		// set using public property pos
		public short mySize;
		public bool isAlive = true;
		///////////////////////////////////////////////////////////////////
		///////////////////////////////////////////////////////////////////
		/// breeding related...
		public geneBankRecord myGeneBankRecord;
		public ergot myChild = null;
		private int numChildren;
		///////////////////////////////////////////////////////////////////
		///////////////////////////////////////////////////////////////////
		/// processor/bias related...
		public short myGenomeUsage = 0;
		public double mySlicerBias = 1.0;
		public int processorDebt = 0;
		///////////////////////////////////////////////////////////////////
		///////////////////////////////////////////////////////////////////
		/// references to my architecture and related structures...
		public Machine myMachine = null;
		public KTL.Node myNode = null;
		public KTL.Node myReaperNode = null;
		public int reaperQueuePosition = 0;
		///////////////////////////////////////////////////////////////////
		///////////////////////////////////////////////////////////////////
		#endregion

		#region constructors
		/// <summary>
		/// Create an embronyic creature, giving it a reference to the Machine it will run.
		/// </summary>
		/// <param name="theMachine">The machine this creature belongs to.</param>
		/// <param name="theSize">The size of the creature.</param>
		public ergot( Machine theMachine, short theSize ) 
		{
			mySize = theSize;
			myMachine = theMachine;
			isAlive = false;
			++myMachine.countAll;
		}

		/// <summary>
		/// Create an embronyic creature, giving it a reference to the Machine on which it will run.
		/// </summary>
		/// <param name="theMachine">The machine this creature belongs to.</param>
		/// <param name="theSize">The size of the creature.</param>
		/// <param name="thePosition">The position in the soup at which to place the creature. Setting the position IGNORES permissions. You could overwrite another creature's instructions!</param>
		public ergot( Machine theMachine, short theSize, int thePosition ) 
		{
			mySize = theSize;
			myMachine = theMachine;
			isAlive = false;
			this.pos = thePosition;
			++myMachine.countAll;
		}

		/// <summary>
		/// Instantiate a creature from an XML stream.
		/// </summary>
		/// <param name="x">The XmlTextReader object to use.</param>
		/// <param name="theMachine">The Machine I will run on.</param>
		/// <param name="theBuilder">The MachineBuilder that is constructing everything.</param>
		public ergot( System.Xml.XmlTextReader x, Machine theMachine)
		{
			myMachine = theMachine;
			isAlive = false;
			this.FromXMLStream( x );
			++myMachine.countAll;
		}

		/// <summary>
		/// Re-use a dead creature to avoid the overhead of garbage collection.
		/// </summary>
		/// <param name="theSize">The dead creature's new size.</param>
		/// <param name="thePosition">The dead creature's new position.</param>
		public void reUse( short theSize, int thePosition )
		{
			mySize = theSize;
			isAlive = false;
			this.pos = thePosition;
			++myMachine.countAll;
		}
		#endregion

		#region Public Member Properties
		/// <summary>
		/// Sets my position and marks my memory space as "used"
		/// Setting the position IGNORES permissions. You could overwrite another creature's instructions!
		/// </summary>
		public int pos
		{
			get
			{
				return myPosition;
			}
			set
			{
				myPosition = value;
				int size = myMachine.size;
				int modu = myMachine.modu;
				int myEnd = (myPosition + mySize)&modu;
				byte[] usage = myMachine.usage;
				if( isAlive )
				{
					for( int i = myPosition; i != myEnd; i = (i+1)&modu ) 
					{
						// adult
						// mark my territory.
						usage[i] = Machine.c_cell_living;		
					}
				}
				else
				{
					for( int i = myPosition; i != myEnd; i = (i+1)&modu ) 
					{
						// embryo
						// mark my territory.
						usage[i] = Machine.c_cell_embryo;
					}
				}
				myMachine.load += mySize;
			}
		}


		/// <summary>
		/// Set or get the number of children this creature has successfully created.
		/// Once this gets past 3, this creature will receive a unique species name.
		/// </summary>
		public int Children
		{
			get
			{
				return numChildren;
			}
			set
			{
				// reward successful breeding by moving me down in the Reaper Queue.
				myMachine.reaperQueue.moveDown(myReaperNode);
				numChildren = value;
			}
		}
		#endregion

		#region fork
		/// <summary>
		/// Re-instantiate ourselves from disk.
		/// We won't bother with genome banking and the species registry unless we
		/// absolutely have to, since the MachineBuilder should have already taken
		/// care of that for us. Whatever species name is provided to this function
		/// will be used, regardless of whether or not my actual genome matches that
		/// species name's corresponding geneBank genome. (cosmic mutations can cause
		/// a genome to drift from its geneBank entry after it is born)
		/// </summary>
		/// <param name="mySpecies">My "forced" species name.</param>
		public void resume( string mySpecies )
		{
			if( ! isAlive )
			{
				return;
			}
			byte[] soup = myMachine.soup;
			int size = myMachine.size;
			int modu = myMachine.modu;
			string myGenome;

			if( myMachine.myGeneBank.ContainsSpecies(mySpecies) )
			{
				// if our species name is recognized, we use the corresponding genome.
				// (even if our genome doesn't exactly match, due to cosmic mutations)
				myGeneBankRecord = myMachine.myGeneBank.getSpeciesRecord(mySpecies);
				myGenome = myGeneBankRecord.myGenome;
			}
			else
			{
				//
				// otherwise we need to find out our true genome and insert ourselves...
				// (more detail in comments below)
				//
				int mCur = myPosition;
				System.Text.StringBuilder myGenomeSB = new System.Text.StringBuilder(mySize<<1,mySize<<1);
				for( short i = 0; i < mySize; ++i )
				{
					myGenomeSB.Append(myMachine.hexinstrs, soup[mCur] << 1, 2 );
					mCur = (mCur+1)&modu;
				}
				myGenome = myGenomeSB.ToString();
				myGeneBankRecord = ((geneBankRecord)myMachine.myGeneBank[myGenome]);
			}
			if( myGeneBankRecord != null )
			{
				myGeneBankRecord.myCurrentPopulation += 1;
			}
			else
			{
				//
				// We ONLY do this if our species was not found in the speciesRegistry.
				// The XML file was modified improperly, so we need to insert a new
				// genome into the geneBank and a new species into the SpeciesRegistry.
				//
				string myNewSpecies = mySpecies;
				if( myNewSpecies == "" ) 
				{
					myNewSpecies = mySize + "/" + myMachine.numSpeciesSighted++;
				}
				myGeneBankRecord = myMachine.myGeneBank.NewGeneBankRecord().initialize(myMachine, myPosition, mySize, myGenome, myNewSpecies);
				myMachine.myGeneBank[myGenome] = myGeneBankRecord;
			}
			//
			object myRef = this;
			if( myNode == null )
			{
				myNode = new KTL.Node(myRef);
			}
			if( myReaperNode == null )
			{
				myReaperNode = new KTL.Node(myRef);
			}
			myMachine.ergotList.push( myNode );

			//myMachine.reaperQueue.unshift( myReaperNode );
			//
			// figure out where to insert in the reaper queue...
			KTL.Node cur = myMachine.reaperQueue.Head;
			while(true)
			{
				if( cur.isNull || ((ergot)cur.Value).reaperQueuePosition >= this.reaperQueuePosition )
				{
					myMachine.reaperQueue.insertNodeAfter( myReaperNode, cur.PrevNode );
					break;
				}
				cur = cur.NextNode;
			}

			//
			++myMachine.countAdult;
		}


		/// <summary>
		/// Instantiates a creature - makes it "alive" - by adding it to the reaper and execution queues.
		/// The creature takes it species name and a few other settings from its parent.
		/// The parent can be null if this creature is being created manually.
		/// </summary>
		/// <param name="myParent">reference to my parent, or null if I'm Adam/Eve.</param>
		public void fork( ergot myParent )
		{
			// temporarily import some of my Machine's variables - for ease of keyboarding :)
			byte[] soup = myMachine.soup;
			byte[] usage = myMachine.usage;
			int size = myMachine.size;
			int modu = myMachine.modu;
			int myEnd = (myPosition + mySize)&modu;

			//
			// genome generation code.
			//
			int mCur = myPosition;
			System.Text.StringBuilder myGenomeSB = new System.Text.StringBuilder(mySize<<1,mySize<<1);
			for( short i = 0; i < mySize; ++i )
			{
				myGenomeSB.Append(myMachine.hexinstrs, soup[mCur] << 1, 2 );
				mCur = (mCur+1)&modu;
			}

			//
			// find out if I'm a new species
			//
			string myGenome = myGenomeSB.ToString();
			myGeneBankRecord = ((geneBankRecord)myMachine.myGeneBank[myGenome]);
			if( myGeneBankRecord == null )
			{
				string myNewSpecies = mySize + "/" + myMachine.numSpeciesSighted++;
				myGeneBankRecord = myMachine.myGeneBank.NewGeneBankRecord().initialize(myMachine, myPosition, mySize, myGenome, myNewSpecies);
				myMachine.myGeneBank[myGenome] = myGeneBankRecord;
			}
			else
			{
				myGeneBankRecord.inc();
			}
			//
			// end genome generation code.
			//

			//
			// set up my registers
			//
			if( myMachine.opForkCopyRegisters && myParent != null )
			{
				ax = myParent.ax;
				bx = myParent.bx;
				cx = myParent.cx;
				dx = myParent.dx;
			}
			else
			{
				ax = 0;
				bx = 0;
				cx = 0;
				dx = 0;
			}
			if( st == null )
			{
				st = new short[ergot.c_stacksize];
			}
			else
			{
				for( int i = 0; i < ergot.c_stacksize; ++i )
				{
					st[i] = 0;
				}
			}
			fl = false;
			sp = 0;
			ip = 0;
			//
			// end register setup
			//

			//
			// set up my references
			//
			object myRef = this;
			if( myNode == null )
			{
				myNode = new KTL.Node(myRef);
			}
			if( myReaperNode == null )
			{
				myReaperNode = new KTL.Node(myRef);
			}
			myMachine.ergotList.push( myNode );
			myMachine.reaperQueue.push( myReaperNode );

			//
			// set up my variables
			//
			myGenomeUsage = 0;
			mySlicerBias = 1.0;
			numChildren = 0;
			processorDebt = 0;
			isAlive = true;

			// set my usage[] sub-space to 1 (it was 3 and 4 while I was an embryo)
			for( int i = myPosition; i != myEnd; i = (i+1)&modu ) 
			{
				// adult
				// mark my territory.
				usage[i] = Machine.c_cell_living;
			}

			++myMachine.countAdult;
		}
		#endregion

		#region kill
		/// <summary>
		/// Kills a creature by removing it from the queues and setting its memory to "unused"
		/// </summary>
		public void kill()
		{
			int size = myMachine.size;
			int modu = myMachine.modu;
			int myEnd = (myPosition + mySize)&modu;
			byte[] usage = myMachine.usage;
			if( isAlive )
			{
				--myMachine.countAdult;
			}
			--myMachine.countAll;
			if( myChild != null && ! myChild.isAlive )
			{
				myChild.kill();
			}
			if( myMachine.opReaperErase )
			{
				// I must clear myself when I die.
				byte[] soup = myMachine.soup;
				for( int i = myPosition; i != myEnd; i = (i+1)&modu ) 
				{
					// un-mark my territory and clear my instructions
					usage[i] = Machine.c_cell_dead;
					soup[i] = 0;
				}
			}
			else
			{
				// I will leave my instructions behind as detritus.
				for( int i = myPosition; i != myEnd; i = (i+1)&modu ) 
				{
					// un-mark my territory.
					usage[i] = Machine.c_cell_dead;
				}
			}
			if( myGeneBankRecord != null )
			{
				myGeneBankRecord.dec();
				if( myGeneBankRecord.myCurrentPopulation < 1 )
				{
					myMachine.myGeneBank.Remove(myGeneBankRecord.myGenome);
				}
			}
			myMachine.load -= mySize;
			if( myNode != null )
			{
				if( myNode.NextNode != null ) 
				{
					// if NextNode != null then we're part of the ergot list.
					// We will now be removed from the list.
					myMachine.ergotList.removeNode(myNode);
				}
				myNode.NextNode = null;
				myNode.PrevNode = null;
			}
			if( myReaperNode != null )
			{
				if( myReaperNode.NextNode != null )
				{
					// if NextNode != null then we're part of the reaper queue.
					// We will now be removed from the list.
					myMachine.reaperQueue.removeNode(myReaperNode);
				}
				myReaperNode.NextNode = null;
				myReaperNode.PrevNode = null;
			}
			myChild = null;
			myGeneBankRecord = null;
			isAlive = false;
			if( myMachine.resurrectionStack_TopIndex < Machine.c_ress_q_size-1 )
			{
				++myMachine.resurrectionStack_TopIndex;
				myMachine.resurrectionStack[myMachine.resurrectionStack_TopIndex] = this;
			}
		}
		#endregion

		#region runFor
		/// <summary>
		/// Run this creature for the specified number of ticks.
		/// (number of instructions to be executed is not necessarily the same as ticks, since
		/// in the future I might make certain operations, such as template searching, more expensive. 
		/// </summary>
		/// <param name="sliceAllocation">the number of ticks for which I should executed.</param>
		public void runFor( int sliceAllocation )
		{
			byte instr;
			Random wire = myMachine.wire;
			int size = myMachine.size;
			int modu = myMachine.modu;
			int myPosPlusSize = myPosition + size;
			int ipos, work, actualTicks;
			short t;
			for( work = 0, actualTicks = 0; work < sliceAllocation && isAlive; ++work, ++actualTicks )
			{
				ipos = (myPosPlusSize+ip)&modu;
				if( ip > 0 && ip < mySize && myMachine.usage[ipos] != Machine.c_cell_executed )
				{
					// mark this part of my address space as being used
					myMachine.usage[ipos] = Machine.c_cell_executed;
					++myGenomeUsage;
				}
				instr = myMachine.soup[ ipos ];
				if( wire.NextDouble() > 0.9 && wire.NextDouble() < myMachine.opErrFlaw )
				{
					// do something strange to the current instruction.
					// what we do is XOR it with either 16, 8, 4, 2, or 1.
					// essentially, we're flipping a random bit during execution.
					instr ^= (byte)(1 << wire.Next(Machine.c_instr_space));
				}
				switch (instr)
				{
					case 0x00: // nop_0
						break;
					case 0x01: // nop_1
						break;
					case 0x02: // nop_2
						break;
					case 0x03: // xor1
						cx ^= 1;
						break;
					case 0x04: // xor2
						cx ^= 2;
						break;
					case 0x05: // shclo
						cx <<= 1;
						break;
					case 0x06: // shcro
						cx >>= 1;
						break;
					case 0x07: // shdlo
						dx <<= 1;
						break;
					case 0x08: // shdro
						dx >>= 1;
						break;
					case 0x09: // azero
						ax = 0;
						break;
					case 0x0A: // bzero
						bx = 0;
						break;
					case 0x0B: // czero
						cx = 0;
						break;
					case 0x0C: // dzero
						dx = 0;
						break;
					case 0x0D: // swapab
						t = ax;
						ax = bx;
						bx = t;
						break;
					case 0x0E: // swapcd
						t = cx;
						cx = dx;
						dx = t;
						break;
					case 0x0F: // xorab_c
						cx = (short)(ax ^ bx);
						break;
					case 0x10: // addac_a
						ax += cx;
						break;
					case 0x11: // subac_a
						ax -= cx;
						break;
					case 0x12: // addab_c
						cx = (short)(ax + bx);
						break;
					case 0x13: // subab_c
						cx = (short)(ax - bx);
						break;
					case 0x14: // addac_c
						cx += ax;
						break;
					case 0x15: // subac_c
						cx = (short)(ax - cx);
						break;
					case 0x16: // addbc_c
						cx += bx;
						break;
					case 0x17: // subbc_c
						cx = (short)(bx - cx);
						break;
					case 0x18: // slt_ab_c
						cx = ax < bx ? (short)1 : (short)0;
						break;
					case 0x19: // seq_ab_c
						cx = ax == bx ? (short)1 : (short)0;
						break;
					case 0x1A: // sne_ab_c
						cx = ax != bx ? (short)1 : (short)0;
						break;
					case 0x1B: // inca
						ax += 1;
						break;
					case 0x1C: // deca
						ax -= 1;
						break;
					case 0x1D: // incb
						bx += 1;
						break;
					case 0x1E: // decb
						bx -= 1;
						break;
					case 0x1F: // incc
						cx += 1;
						break;
					case 0x20: // decc
						cx -= 1;
						break;
					case 0x21: // incd
						dx += 1;
						break;
					case 0x22: // decd
						dx -= 1;
						break;
					case 0x23: // pushax
						st[sp] = ax;
						sp = ergot.c_spinc[sp];
						break;
					case 0x24: // pushbx
						st[sp] = bx;
						sp = ergot.c_spinc[sp];
						break;
					case 0x25: // pushcx
						st[sp] = cx;
						sp = ergot.c_spinc[sp];
						break;
					case 0x26: // pushdx
						st[sp] = dx;
						sp = ergot.c_spinc[sp];
						break;
					case 0x27: // popax
						sp = ergot.c_spdec[sp];
						ax = st[sp];
						break;
					case 0x28: // popbx
						sp = ergot.c_spdec[sp];
						bx = st[sp];
						break;
					case 0x29: // popcx
						sp = ergot.c_spdec[sp];
						cx = st[sp];
						break;
					case 0x2A: // popdx
						sp = ergot.c_spdec[sp];
						dx = st[sp];
						break;
					case 0x2B: // if_cz
						if( cx != 0 ) { ++ip; }
						break;
					case 0x2C: // if_cz4
						if( cx != 0 ) { ip += 4; }
						break;
					case 0x2D: // if_cz8
						if( cx != 0 ) { ip += 8; }
						break;
					case 0x2E: // if_dz
						if( dx != 0 ) { ++ip; }
						break;
					case 0x2F: // if_dz4
						if( dx != 0 ) { ip += 4; }
						break;
					case 0x30: // if_dz8
						if( dx != 0 ) { ip += 8; }
						break;
					case 0x31: // jump
						work += 49;
						myMachine.doTwoWayADDR( this, true, false );
						break;
					case 0x32: // jmpf
						work += 24;
						myMachine.doForwardADDR( this, true, false );
						break;
					case 0x33: // jmpb
						work += 24;
						myMachine.doBackwardADDR( this, true, false );
						break;
					case 0x34: // call
						work += 49;
						myMachine.doTwoWayADDR( this, true, true );
						break;
					case 0x35: // ret
						sp = ergot.c_spdec[sp];
						ip = st[sp];
						break;
					case 0x36: // addr
						work += 49;
						myMachine.doTwoWayADDR( this, false, false );
						break;
					case 0x37: // adrf
						work += 24;
						myMachine.doForwardADDR( this, false, false );
						break;
					case 0x38: // adrb
						work += 24;
						myMachine.doBackwardADDR( this, false, false );
						break;
					case 0x39: // mov_ab
						bx = ax;
						break;
					case 0x3A: // mov_cd
						dx = cx;
						break;
					case 0x3B: // mov_ipc
						cx = ip;
						break;
					case 0x3C: // mov_iab
						work += 1;
						int axpos = (myPosPlusSize+ax)&modu;
						int bxpos = (myPosPlusSize+bx)&modu;

						if( myChild == null )
						{
							// I do not have a child.
							if( myMachine.usage[axpos] != Machine.c_cell_dead
								&& (ax < 0 || ax >= mySize) )
							{
								// the cell at offset ax is in use,
								// and (ax is not in my address space)
								// ABORT.
								fl = true;
								break;
							}
						}
						else
						{
							// I have a child.
							int myChildOffset = myChild.ax;	// ( myChild.ax is offset, set by FindEmptySpace )
							int myChildEndOffset = myChildOffset + myChild.mySize;
							if( myMachine.usage[axpos] == Machine.c_cell_dead )
							{
								;
							}
							else if( ax >= 0 && ax < mySize )
							{
								;
							}
							else if( ax >= myChildOffset && ax < myChildEndOffset )
							{
								if( myMachine.usage[axpos] != Machine.c_cell_embwri )
								{
									++myChild.bx;
									myMachine.usage[axpos] = Machine.c_cell_embwri;
								}
							}
							else
							{
								fl = true;
								break;
							}
						}
						if( wire.NextDouble() > 0.99 && wire.NextDouble() < myMachine.opErrCopy )
						{
							// do something strange to the instruction being copied.
							// what we do is XOR it with either 32, 16, 8, 4, 2, or 1.
							// essentially, we're flipping a random bit.
							myMachine.soup[axpos] = (byte)(myMachine.soup[bxpos] ^ (1 << wire.Next(Machine.c_instr_space)));
						}
						else
						{
							myMachine.soup[axpos] = myMachine.soup[bxpos];
						}
						break;
					case 0x3D: // malloc
						work += 74;
						if( myChild != null || dx < 12 || dx > (mySize * 3) )
						{
							// if I have a child already, or the size for the child (dx)
							// is too small or too large, then set error flag and do nothing.
							fl = true;
						}
						else
						{
							myChild = myMachine.findEmptySpace( myPosition, ax, dx );
							if( myChild == null )
							{
								// if my child creation failed, set error flag and do nothing.
								fl = true;
							}
							else
							{
								// child creation succeeded. set parent's ax to child offset
								// (findEmptySpace set the child's ax register to its offset for us)
								ax = myChild.ax;
								myChild.bx = 0;
							}
						}
						break;
					case 0x3E: // fork
						work += 7;
						if( myChild == null || ((myChild.bx << 1 ) < myChild.mySize) )
						{
							this.fl = true;
						}
						else
						{
							this.Children += 1;
							if( myMachine.opSoupSliceLeanBias )
							{
								mySlicerBias = myGenomeUsage * 1.5;
								mySlicerBias /= mySize;
							}
							myChild.fork(this);
							myChild = null;
						}
						break;
					case 0x3F: // flag
						if( myMachine.opErrFlagInstr )
						{
							processorDebt += 50;
							fl = true;
						}
						break;
					default:
						// what.
						fl = true;
						break;
				} // end switch
				if( fl )
				{
					// punish errors by moving me up in the Reaper Queue.
					myMachine.reaperQueue.moveUp(myReaperNode);
					fl = false;
				}
				if( wire.NextDouble() < myMachine.opErrFlaw )
				{
					// make the ip stutter ( +1, -1, or stay the same )
					ip += (short)(wire.Next(3) - 1);
				}
				++ip;
			} // end for loop
			myMachine.numTicks += (uint)actualTicks;
			processorDebt += work - sliceAllocation;
			myMachine.amntWork += (uint)sliceAllocation;
		} // end function
		#endregion

		#region FromXMLStream
		/// <summary>
		/// Import a creature from an XML stream.
		/// </summary>
		/// <param name="x">The XmlTextReader to use.</param>
		/// <param name="theBuilder">A reference to the MachineBuilder that is taking care of everything.</param>
		public void FromXMLStream( XmlTextReader x )
		{
			while( x.Name == "" )
			{
				x.Read();
			}
			if( x.Name.ToLower() != "ergot" && x.Name.ToLower() != "embryo" )
			{
				// not the correct type of grouping.
				return;
			}
			string nodeName = "";
			string mySpecies = "";
			bool keepReading = true;

			while( keepReading && x.Read() )
			{
				switch (x.NodeType)
				{
					case XmlNodeType.Element:
						nodeName = x.Name.ToLower();
					switch (nodeName)
					{
						case "genome":
							break;		// for the moment, we don't do anything about genome.
						case "mychild":
							x.Read();	// advance past the myChild tag.
							myChild = new ergot(x, myMachine );
							break;
						default:
							break;
					}
						break;
					case XmlNodeType.Text:
					switch (nodeName)
					{
						case "ax":
							try
							{
								ax = System.Int16.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "bx":
							try
							{
								bx = System.Int16.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "cx":
							try
							{
								cx = System.Int16.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "dx":
							try
							{
								dx = System.Int16.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "ip":
							try
							{
								ip = System.Int16.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "sp":
							try
							{
								sp = System.Byte.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "fl":
							try
							{
								fl = System.Boolean.Parse( x.Value );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "st":
							try
							{
								st = new short[ergot.c_stacksize];
								string tmp1 = System.Text.RegularExpressions.Regex.Replace( x.Value, "\\s+", "" );
								string[] tmp = tmp1.Split(",".ToCharArray(),ergot.c_stacksize);
								for( int i = 0; i < ergot.c_stacksize; ++i )
								{
									st[i] = System.Int16.Parse( tmp[i], myMachine.intStyle );
								}
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "isalive":
							try
							{
								isAlive = System.Boolean.Parse( x.Value );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "myposition":
							try
							{
								myPosition = System.Int32.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "mysize":
							try
							{
								mySize = System.Int16.Parse( x.Value, myMachine.intStyle );
								myMachine.load += mySize;
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "myspecies":
							mySpecies = x.Value;
							break;
						case "mycurrentpopulation":
							break;
						case "mypeak":
							break;
						case "mytotallivinganddead":
							break;
						case "numchildren":
							try
							{
								numChildren = System.Int32.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "mygenomeusage":
							try
							{
								myGenomeUsage = System.Int16.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "myslicerbias":
							try
							{
								mySlicerBias = System.Double.Parse( x.Value, myMachine.dblStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "processordebt":
							try
							{
								processorDebt = System.Int32.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						case "reaperqueueposition":
							try
							{
								reaperQueuePosition = System.Int32.Parse( x.Value, myMachine.intStyle );
							}
							catch (Exception e)
							{
								myMachine.ioError(nodeName,e);
							}
							break;
						default:
							break;
					}
						break;
					case XmlNodeType.EndElement:
						if( x.Name.ToLower() == "ergot" || x.Name.ToLower() == "embryo" )
						{
							keepReading = false;
							break;
						}
						nodeName = "";
						break;
					default:
						break;
				} // switch
			} // while
			this.resume( mySpecies );
		} // function
		#endregion

		#region ToXMLStream
		/// <summary>
		/// Output the entire contents of this creature, and its embryonic daughter,
		/// to the specified XML writer stream.
		/// </summary>
		/// <param name="x">An XmlTextWriter object.</param>
		public void ToXMLStream( System.Xml.XmlTextWriter x )
		{
			if( isAlive )
			{
				x.WriteStartElement("Ergot");
			}
			else
			{
				x.WriteStartElement("Embryo");
			}

			if(true)
			{
				x.WriteElementString( "ax", ax.ToString() );
				x.WriteElementString( "bx", bx.ToString() );
				x.WriteElementString( "cx", cx.ToString() );
				x.WriteElementString( "dx", dx.ToString() );
				x.WriteElementString( "ip", ip.ToString() );
				x.WriteElementString( "sp", sp.ToString() );
				x.WriteElementString( "fl", fl.ToString() );
				if( st != null )
				{
					x.WriteStartElement( "st" );
					if(true)
					{
						for( int i = 0; i < ergot.c_stacksize; ++i )
						{
							x.WriteString( st[i].ToString() );
							if( i < ergot.c_stacksize - 1 )
							{
								x.WriteString( "," );
							}
						}
					}
					x.WriteEndElement();
				}
				x.WriteElementString( "isAlive", isAlive.ToString() );
				x.WriteElementString( "myPosition", myPosition.ToString() );
				x.WriteElementString( "mySize", mySize.ToString() );
				if( myGeneBankRecord != null )
				{
					x.WriteElementString( "mySpecies", myGeneBankRecord.mySpeciesName.ToString() );
					x.WriteElementString( "myCurrentPopulation", myGeneBankRecord.myCurrentPopulation.ToString() );
					x.WriteElementString( "myTotalLivingAndDead", myGeneBankRecord.myCount.ToString() );
				}
				x.WriteElementString( "numChildren", numChildren.ToString() );
				x.WriteElementString( "myGenomeUsage", myGenomeUsage.ToString() );
				x.WriteElementString( "mySlicerBias", mySlicerBias.ToString() );
				x.WriteElementString( "processorDebt", processorDebt.ToString() );
				x.WriteElementString( "reaperQueuePosition", reaperQueuePosition.ToString() );
				x.WriteStartElement( "genome" );
				if(true)
				{
					// DEBUG: GENOME PRINTING
					int myEnd = (myPosition+mySize)&myMachine.modu;
					int curPos = myPosition;
					for( int i = 0; i < mySize; ++i, curPos = (curPos+1)&myMachine.modu )
					{
						x.WriteStartElement( "cell" );
						x.WriteAttributeString( "", "index", null, i.ToString("D3") );
						x.WriteAttributeString( "", "dec", null, myMachine.soup[curPos].ToString("D2") );
						x.WriteAttributeString( "", "hex", null, myMachine.soup[curPos].ToString("X2") );
						x.WriteString(myMachine.instrs[myMachine.soup[curPos]]);
						x.WriteEndElement();
					}
				}
				x.WriteEndElement();
				if( myChild != null )
				{
					x.WriteStartElement("myChild");
					myChild.ToXMLStream(x);
					x.WriteEndElement();
				}
			}
			x.WriteEndElement();
		}
		#endregion


	}	// end class ergot
}	// end namespace ERGOT
